

function Service() {
  this.register();
}

Service.prototype={
  get wrappedJSObject() {
    return this;
  }
,
  QueryInterface: function(iid) {
     this.queryInterfaceSupport(iid,SERVICE_IIDS);
     return this;
  }
,
  // nsIObserver implementation 
  observe: function(subject, topic, data) {
    if(this.customObserve) {
      if(this.customObserve(subject,topic,data)) {
        return;
      }
    }
    if(subject instanceof Components.interfaces.nsIPrefBranchInternal
      && this.syncPrefs) {
      this.syncPrefs(subject,data);
    } else {
      switch(topic) {
        case "xpcom-shutdown":
          this.unregister();
          break;
        case "profile-before-change": 
          this.cleanup();
          if(this._queue) this._queue.running=false;
          break;
        case "profile-after-change":
          this.init();
          break;
      }
    }
  }
,  
  register: function() {
    const osvr=Components.classes['@mozilla.org/observer-service;1'].getService(
    Components.interfaces.nsIObserverService);
    osvr.addObserver(this,"profile-before-change",false);
    osvr.addObserver(this,"xpcom-shutdown",false);
    osvr.addObserver(this,"profile-after-change",false);
    if(this.customRegister) this.customRegister();
  }
,
  unregister: function() {
    const osvr=Components.classes['@mozilla.org/observer-service;1'].getService(
      Components.interfaces.nsIObserverService);
    osvr.removeObserver(this,"profile-before-change");
    osvr.removeObserver(this,"xpcom-shutdown");
    osvr.removeObserver(this,"profile-after-change");
    if(this.customUnregister) this.customUnregister();
  }
,
  _inited: false,
  init: function() {
    if(this._inited) return;
    this._inited=true;
    
    const prefserv=this.prefService=Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService)
    const PBI=Components.interfaces.nsIPrefBranchInternal;
    this.prefs=prefserv.getBranch(SERVICE_BRANCH_NAME).QueryInterface(PBI);
    if(this.customInit) this.customInit();
  }
,
  willBeUninstalled: function() {
   if(this.uninstalling) return true;
   try {
     const RDFService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
               .getService(Components.interfaces.nsIRDFService);
     const container = Components.classes["@mozilla.org/rdf/container;1"]
               .getService(Components.interfaces.nsIRDFContainer);
     const extensionDS= Components.classes["@mozilla.org/extensions/manager;1"]
          .getService(Components.interfaces.nsIExtensionManager).datasource;
     var root = RDFService.GetResource("urn:mozilla:extension:root");
     const nameArc = RDFService.GetResource("http://www.mozilla.org/2004/em-rdf#name");
     const toBeUninstalledArc = RDFService.GetResource("http://www.mozilla.org/2004/em-rdf#toBeUninstalled");
     const toBeDisabledArc=RDFService.GetResource("http://www.mozilla.org/2004/em-rdf#toBeDisabled");
     container.Init(extensionDS,root);
    
     var found = false;
     var elements = container.GetElements();
     var element,name,target;
     while (elements.hasMoreElements()) {
      element = elements.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
     
  
      if((target=extensionDS.GetTarget(element, nameArc ,true)) && 
          target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value==SERVICE_NAME
        && 
        (
          (target = extensionDS.GetTarget(element, toBeUninstalledArc,true))
           && target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "true"
          ||
          (target = extensionDS.GetTarget(element, toBeDisabledArc,true))
          && target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "true"
        )  
      ) {
         return this.uninstalling=true;
        }         
     }
   } catch(ex) {} // quick and dirty work-around for Mozilla ;)
   return this.uninstalling=false;
  }
,
  cleanup: function() {
    this.cleanupIfUninstalling();
  }
,
  cleanupIfUninstalling: function() {
    if(this.willBeUninstalled()) this.uninstallJob();
    return this.uninstalling;
  }
,
  uninstallJob: function() {
  }
,
  getPref: function(name,def) {
    const IPC=Components.interfaces.nsIPrefBranch;
    const prefs=this.prefs;
    try {
      switch(prefs.getPrefType(name)) {
        case IPC.PREF_STRING:
          return prefs.getCharPref(name);
        case IPC.PREF_INT:
          return prefs.getIntPref(name);
        case IPC.PREF_BOOL:
          return prefs.getBoolPref(name);
      }
    } catch(e) {}
    return def;
  }
,
  setPref: function(name,value) {
    const prefs=this.prefs;
    switch(typeof(value)) {
      case "string":
          prefs.setCharPref(name,value);
          break;
      case "boolean":
        prefs.setBoolPref(name,value);
        break;
      case "number":
        prefs.setIntPref(name,value);
        break;
      default:
        throw new Error("Unsupported type "+typeof(value)+" for preference "+name);
    }
  }
,
  _sound: null,
  playSound: function(url) {
    if(this.getPref("sound",true)) {
      var sound=this._sound;
      if(sound==null) {
        sound=Components.classes["@mozilla.org/sound;1"].createInstance(Components.interfaces.nsISound);
        sound.init();
      }
      try {
        sound.play(this.ios.newURI(url,null,null));
      } catch(ex) {
        dump(ex);
      }
    }
  }
,
  readFile: function(file) {
    const cc=Components.classes;
    const ci=Components.interfaces;  
    const is = cc["@mozilla.org/network/file-input-stream;1"].createInstance(
          ci.nsIFileInputStream );
    is.init(file ,0x01, 0400, null);
    const sis = cc["@mozilla.org/scriptableinputstream;1"].createInstance(
      ci.nsIScriptableInputStream );
    sis.init(is);
    const res=sis.read(sis.available());
    is.close();
    return res;
  }
,
  writeFile: function(file, content) {
    const cc=Components.classes;
    const ci=Components.interfaces;
    const unicodeConverter = cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(
    ci.nsIScriptableUnicodeConverter);
    unicodeConverter.charset = "UTF-8";
    content=unicodeConverter.ConvertFromUnicode(content);
    const os=cc["@mozilla.org/network/file-output-stream;1"].createInstance(
      ci.nsIFileOutputStream);
    os.init(file, 0x02 | 0x08 | 0x20,0664,0);
    os.write(content,content.length);
    os.close();
  }
,
  get prompter() {
    return Components.classes["@mozilla.org/embedcomp/prompt-service;1"
          ].getService(Components.interfaces.nsIPromptService);
  }
,
  queryInterfaceSupport: function(iid,iids) { 
    xpcom_checkInterfaces(iid, iids, Components.results.NS_ERROR_NO_INTERFACE);
  }
,
// CUSTOM CODE
 
  createValidator: function() {
    return  Components.classes['@maone.net/csevalid;1'].createInstance(
      Components.interfaces.nsICsevalid);
  }
,
  _safeOp: function(op) {
     var c=null;
     var res=null;
     var exception=null;
     try {
       c=this.createValidator();
       res = op(c,this);
     } catch(ex) {
       exception=new Error("Exception "+ex+" calling "+op+" on cse validator.\n");
     }
     if(c) {
       c.dispose();
       delete c;
     }
     if(exception) throw exception;
     return res;
  }
,
  checkEngine: function() {
    if(! ("engineCheck" in this) ) {
      if(this.getPref("engineFail",false)) { // for debug purpose
        this.engineCheck=false;
      } else {
        this.engineCheck=true;
        const serv=this;
        this.validateRunner.run(function() {
            serv._safeOp(function(c,t) { 
              t.engineCheck=c.initialized;
            })
          });
      }
    }
    return this.engineCheck;
  }
,
  get timestamp() {
    return new Date().getTime();
  }
,
  _lastConfChange: null,
  get lastConfChange() {
    return this._lastConfChange==null?this._lastConfChange=this.timestamp:this._lastConfChange;
  }
,
  confChanged: function() {
    this._lastConfChange=this.timestamp;
  }
,
  isObsolete: function(results) {
    return ("timestamp" in results) && results.timestamp<this.lastConfChange;
  }
,
  showOptionsDialog: function() {
    this._safeOp(function(c) {
      return c.showOptionsDialog();
    });
    this.confChanged();
  }
,
  readMessageType: function(msgid) {
    return this._safeOp(function(c) { 
      return c.getType(msgid); 
    });
  }
,
  writeMessageType: function(msgid,typeNum) {
     this._safeOp(function(c) { 
       c.setType(msgid,typeNum);
     });
     this.confChanged();
  }
,
  readMessageEnabled: function(msgid) {
    return this._safeOp(function(c) { 
      return c.isEnabled(msgid); 
    });
  }
,
  writeMessageEnabled: function(msgid,b) {
     this._safeOp(function(c) { 
       c.setEnabled(msgid,b);
     });
     this.confChanged();
  }
,
  trim2hash: function(url) {
    const pos=url.lastIndexOf("#");
    return pos<0?url:url.substring(0,pos);
  }
,
  areURIsEqual: function(u1,u2) {
    return this.trim2hash(u1)==this.trim2hash(u2);
  }
,
  _validateRunner: null,
  get validateRunner() {
    if(!this._validateRunner) {
      const async=this.getPref("async",false);
      // dump("CSE using "+(async?"an async":"a sync")+" Runner\n");
      this._validateRunner=
        async?new AsyncRunner(new LastRunsStarter())
          :new SyncRunner();
    }
    return this._validateRunner;
  },
  validate: function(html,results) {   
    const service=this;
    
    const mainJob=function() {
      var exception=null;
      var validator=null;
      
      try {
        validator=service.createValidator();
        const messagesCount=validator.validate(results.lines.length,results.lines)
        delete results.lines;
        results.messagesCount=messagesCount;
        if(messagesCount>=0) {
          results.errors=validator.getMessageCount(validator.TYPE_ERROR);
          results.warnings=validator.getMessageCount(validator.TYPE_WARNING);
          results.messages=null; // remove this if you don't want early message fetching
          if('messages' in results) {
            // fetch messages
            var arg={
              text:{},
              cat: {},
              type:{},
              line:{},
              charpos:{},
              length:{},
              flags:{},
              id: {}
            };
            const messages=new Array(messagesCount);
            var message;
            for(var j=0; j<messagesCount; j++) {
              validator.getMessageAt(j,
                arg.text,arg.cat,arg.type,arg.line,arg.charpos,arg.length,arg.flags,arg.id);
     
              message={};
              for(p in arg) {
                message[p]=arg[p].value;
              }
              messages[j]=message;
            }
            results.messages=messages;
          } 
        }
      } catch(ex) {
        exception=ex;
      }
      if(validator!=null) {
        validator.dispose();
        delete validator;
      }
      if(exception!=null) throw exception;
    }
    
    const postJob=function() { 
      try {
        results.working=false;
        if(results.token) {
          this.dump(results.token+" VALIDATION DONE.\n\n");
        }
        if(results.callback) {
          const callback=results.callback;
          delete results.callback;
          callback(results);
        }
      } catch(ex) {
        dump(ex+"\n");
      }
    }

    results.errors=0;
    results.warnings=0;
    results.timestamp=this.lastConfChange;
    if(html==null || html.length==0) {  
      postJob();
      return;
    }
    
    results.working=true;
    results.lines=html.split(/\n/);
    delete html;
    
    if(results.token) {
      this.dumpWithStack(results.token+" PERFORMING VALIDATION - stack:");
    }
    
    this.validateRunner.run(mainJob,postJob);
  }
,
  dumpWithStack: function (msg) {
    dump(msg);
    dump("\n");
    for(var stack=Components.stack; stack=stack.caller;) {
      dump(" in ");
      dump(stack);
      dump("\n");
    }
    dump("\n");
  } 
};


// Thread utility objects - Copyright (c) 2005 Giorgio Maone (InformAction)

const DEF_THREAD_FACTORY = {
  createThread: function(runnable) {
    const cc=Components.classes;
    const ci=Components.interfaces;
    const thread=cc['@mozilla.org/thread;1'].createInstance(ci.nsIThread);
    if(runnable) {
      runnable.thread=thread;
      thread.init(runnable, 0,  
            ci.nsIThread.PRIORITY_NORMAL, ci.nsIThread.SCOPE_GLOBAL, ci.nsIThread.STATE_JOINABLE);
    }
    return thread;
  }
,
  get currentThread() {
    return DEF_THREAD_FACTORY.createThread().currentThread;
  }
}

function SimpleStarter(threadFactory) {
  this.threadFactory=threadFactory?threadFactory:DEF_THREAD_FACTORY;
}

SimpleStarter.prototype = {
  start: function(runnable) {
    this.threadFactory.createThread(runnable);
  }
}

function QueueStarter(threadFactory) {
  this.threadFactory=threadFactory?threadFactory:DEF_THREAD_FACTORY;
  this.queue=new RunnableQueue(this.threadFactory);
}
QueueStarter.prototype = {
  start: function(runnable) {
    this.queue.enqueue(runnable);
  }
}

function LastRunsStarter(threadFactory) {
  this.threadFactory=threadFactory?threadFactory:DEF_THREAD_FACTORY;
  this.threadFactory.createThread(this);
}
LastRunsStarter.prototype = {
  current: null,
  running: true,
  run: function() {
    while(this.running) {
      if(this.current && this.current.run) {
        this.current.run();
        this.current=null;
      }
      this.thread.sleep(100);
    }
    delete this.thread;
  }
,
  start: function(runnable) {
    this.current=runnable;
  }
}

function SyncRunner() {}
SyncRunner.prototype={
  run: function(mainJob,postJob) {
    try {
      mainJob();
    } catch(ex) {
       dump("SyncRunner mainJob error: "+ex+"\n");
    }
    try {
      if(postJob) postJob();
    } catch(ex) {
      dump("SyncRunner postJob error: "+ex+"\n");
    }
  }
}

function AsyncRunner(starter) {
  if(starter) this.starter=starter;
}

AsyncRunner.prototype={
  starter: new SimpleStarter()
,  
  run: function(asyncJob,postJob) {
    const cc=Components.classes;
    const ci=Components.interfaces;
  
    var postRunnable=null;
    const asyncError={ value: null };
    if(postJob) {
      postRunnable = cc["@mozilla.org/xpcomproxy;1"].getService(ci.nsIProxyObjectManager
        ).getProxyForObject(
          cc["@mozilla.org/event-queue-service;1"].getService(ci.nsIEventQueueService
              ).getSpecialEventQueue(ci.nsIEventQueueService.UI_THREAD_EVENT_QUEUE),
          ci.nsIRunnable, 
          {
            QueryInterface: function(iid) {
              if(iid.equals(ci.nsISupport) || iid.equals(ci.nsIRunnable)) {
               return this;
              }
              throw Components.results.NS_ERROR_NO_INTERFACE;
            },
            run: function() { 
              if(asyncError.value) {
                dump("AsyncRunner mainJob error: "+asyncError.value+"\n");
              }
              try {
                postJob();
              } catch(ex) {
                dump("AsyncRunner postJob error: "+ex+"\n");
              }
            }
          }
        , 5); // 5 == PROXY_ALWAYS | PROXY_SYNC
    }
    
    const asyncRunnable={
      run: function() { 
        try {
          asyncJob();
        } catch(ex) {
          asyncError.value=ex;
        }
        if(postRunnable) postRunnable.run();
      }
    };
    
    this.starter.start(asyncRunnable);
  }  
}

function RunnableQueue(threadFactory) {
  this.current=this.first=new Object();
  if(!threadFactory) threadFactory=new AsyncRunner();
  threadFactory.createThread(this);
}
RunnableQueue.prototype={
  first: null,
  current: null,
  enqueue: function(r) {
    this.current=this.current.next=r;
  },
  running: true,  
  run: function() {
    var r=this.first;
    while(this.running) {
      if(!r) return;
      if(r.run) {
        try {
          r.run();
          r.run=null;
        } catch(ex) {
          //dump(ex);
        }
      }
      if(r.next) {
        r=r.next;
      }
      this.thread.sleep(100);
    }
    delete this.thread;
  }
}



// XPCOM Scaffolding code

// component defined in this file

const SERVICE_NAME="HTML Validator (based on CSE HTML Validator)";
const SERVICE_CID =
    Components.ID("{5d1376d6-cf9c-4bb9-ac90-bcddde37cdaa}");
const SERVICE_CTRID =
    "@maone.net/csevalid-service;1";
const SERVICE_BRANCH_NAME="csevalid.";    
const SERVICE_CONSTRUCTOR=Service;

// interfaces implemented by this component
const SERVICE_IIDS = 
[ 
Components.interfaces.nsIObserver,
Components.interfaces.nsISupports,
Components.interfaces.nsISupportsWeakReference
];

// Factory object
const SERVICE_FACTORY = {
  _instance: new SERVICE_CONSTRUCTOR(),
  createInstance: function (outer, iid) {
    if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;

    xpcom_checkInterfaces(iid,SERVICE_IIDS,Components.results.NS_ERROR_INVALID_ARG);
    // kept this for flexibility sake, but we're really adopting an
    // early instantiation and late init singleton pattern
    return this._instance==null?this._instance=new SERVICE_CONSTRUCTOR():this._instance;
  }
};

function xpcom_checkInterfaces(iid,iids,ex) {
  for(var j=iids.length; j-- >0;) {
    if(iid.equals(iids[j])) return true;
  }
  throw ex;
}

// Module

var Module = new Object();
Module.firstTime=true;
Module.registerSelf = function (compMgr, fileSpec, location, type) {
  if(this.firstTime) {
   
    debug("*** Registering "+SERVICE_CTRID+".\n");
    
    compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar
      ).registerFactoryLocation(SERVICE_CID,
      SERVICE_NAME,
      SERVICE_CTRID, 
      fileSpec,
      location, 
      type);
      
    Components.classes['@mozilla.org/categorymanager;1'].getService(
      Components.interfaces.nsICategoryManager
     ).addCategoryEntry("app-startup",
        SERVICE_NAME, "service," + SERVICE_CTRID, true, true, null);
      
    this.firstTime=false;
  } 
}

Module.unregisterSelf = function(compMgr, fileSpec, location) {
  compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar
    ).unregisterFactoryLocation(SERVICE_CID, fileSpec);
  Components.classes['@mozilla.org/categorymanager;1'].getService(
      Components.interfaces.nsICategoryManager
     ).deleteCategoryEntry("app-startup",SERVICE_NAME, true);
}

Module.getClassObject = function (compMgr, cid, iid) {
  if(cid.equals(SERVICE_CID))
    return SERVICE_FACTORY;

  if (!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  
  throw Components.results.NS_ERROR_NO_INTERFACE;
    
}

Module.canUnload = function(compMgr) {
  return true;
}

// entrypoint
function NSGetModule(compMgr, fileSpec) {
  return Module;
}



